@@ -5,18 +5,16 @@
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution.
 from __future__ import absolute_import, division, unicode_literals
+
 import decimal
-import warnings
 import sys
 import types
+import warnings
 from itertools import chain, islice
 
-from . import compat
-from . import util
-from . import tags
-from . import handlers
+from . import compat, handlers, tags, util
 from .backend import json
-from .compat import numeric_types, string_types, PY3, PY2
+from .compat import PY2, PY3, numeric_types, string_types
 
 
 def encode(
@@ -39,9 +37,17 @@ def encode(
 ):
     """Return a JSON formatted representation of value, a Python object.
 
-    :param unpicklable: If set to False then the output will not contain the
+    :param unpicklable: If set to ``False`` then the output will not contain the
         information necessary to turn the JSON data back into Python objects,
-        but a simpler JSON stream is produced.
+        but a simpler JSON stream is produced. It's recommended to set this
+        parameter to ``False`` when your code does not rely on two objects
+        having the same ``id()`` value, and when it is sufficient for those two
+        objects to be equal by ``==``, such as when serializing sklearn
+        instances. If you experience (de)serialization being incorrect when you
+        use numpy, pandas, or sklearn handlers, this should be set to ``False``.
+        If you want the output to not include the dtype for numpy arrays, add
+        ``jsonpickle.register(numpyp.generic,
+         UnpicklableNumpyGenericHandler, base=True)`` before your pickling code.
     :param max_depth: If set to a non-negative integer then jsonpickle will
         not recurse deeper than 'max_depth' steps into the object.  Anything
         deeper than 'max_depth' is represented using a Python repr() of the
@@ -130,6 +136,32 @@ def encode(
     )
 
 
+def _in_cycle(obj, objs, max_reached, make_refs):
+    """Detect cyclic structures that would lead to infinite recursion"""
+    return (
+        (max_reached or (not make_refs and id(obj) in objs))
+        and not util.is_primitive(obj)
+        and not util.is_enum(obj)
+    )
+
+
+def _mktyperef(obj):
+    """Return a typeref dictionary
+
+    >>> _mktyperef(AssertionError) == {'py/type': 'builtins.AssertionError'}
+    True
+
+    """
+    return {tags.TYPE: util.importable_name(obj)}
+
+
+def _wrap_string_slot(string):
+    """Converts __slots__ = 'a' into __slots__ = ('a',)"""
+    if isinstance(string, string_types):
+        return (string,)
+    return string
+
+
 class Pickler(object):
     def __init__(
         self,
@@ -222,6 +254,16 @@ class Pickler(object):
     def _getref(self, obj):
         return {tags.ID: self._objs.get(id(obj))}
 
+    def _flatten(self, obj):
+        if self.unpicklable and self.make_refs:
+            result = self._flatten_impl(obj)
+        else:
+            try:
+                result = self._flattened[id(obj)]
+            except KeyError:
+                result = self._flattened[id(obj)] = self._flatten_impl(obj)
+        return result
+
     def flatten(self, obj, reset=True):
         """Takes an object and returns a JSON-safe representation of it.
 
@@ -254,15 +296,20 @@ class Pickler(object):
             self.reset()
         return self._flatten(obj)
 
-    def _flatten(self, obj):
-        if self.unpicklable and self.make_refs:
-            result = self._flatten_impl(obj)
-        else:
+    def _flatten_file(self, obj):
+        """
+        Special case file objects
+        """
+        assert not PY3 and isinstance(obj, types.FileType)
+        return None
+
+    def _flatten_bytestring(self, obj):
+        if PY2:
             try:
-                result = self._flattened[id(obj)]
-            except KeyError:
-                result = self._flattened[id(obj)] = self._flatten_impl(obj)
-        return result
+                return obj.decode('utf-8')
+            except UnicodeDecodeError:
+                pass
+        return {self._bytes_tag: self._bytes_encoder(obj)}
 
     def _flatten_impl(self, obj):
         #########################################
@@ -271,14 +318,13 @@ class Pickler(object):
         if PY2 and isinstance(obj, types.FileType):
             return self._flatten_file(obj)
 
-        if util.is_bytes(obj):
+        if type(obj) is bytes:
             return self._flatten_bytestring(obj)
 
-        if util.is_primitive(obj):
-            return obj
-
         # Decimal is a primitive when use_decimal is True
-        if self._use_decimal and isinstance(obj, decimal.Decimal):
+        if type(obj) in util.PRIMITIVES or (
+            self._use_decimal and isinstance(obj, decimal.Decimal)
+        ):
             return obj
         #########################################
 
@@ -288,6 +334,11 @@ class Pickler(object):
     def _max_reached(self):
         return self._depth == self._max_depth
 
+    def _pickle_warning(self, obj):
+        if self.warn:
+            msg = 'jsonpickle cannot pickle %r: replaced with None' % obj
+            warnings.warn(msg)
+
     def _flatten_obj(self, obj):
         self._seen.append(obj)
 
@@ -319,38 +370,41 @@ class Pickler(object):
     def _list_recurse(self, obj):
         return [self._flatten(v) for v in obj]
 
+    def _flatten_function(self, obj):
+        if self.unpicklable:
+            data = {tags.FUNCTION: util.importable_name(obj)}
+        else:
+            data = None
+
+        return data
+
     def _get_flattener(self, obj):
-
-        list_recurse = self._list_recurse
-
-        if util.is_list(obj):
+        if type(obj) in (list, dict):
             if self._mkref(obj):
-                return list_recurse
+                return (
+                    self._list_recurse if type(obj) is list else self._flatten_dict_obj
+                )
             else:
                 self._push()
                 return self._getref
 
         # We handle tuples and sets by encoding them in a "(tuple|set)dict"
-        if util.is_tuple(obj):
+        elif type(obj) in (tuple, set):
             if not self.unpicklable:
-                return list_recurse
-            return lambda obj: {tags.TUPLE: [self._flatten(v) for v in obj]}
+                return self._list_recurse
+            return lambda obj: {
+                tags.TUPLE
+                if type(obj) is tuple
+                else tags.SET: [self._flatten(v) for v in obj]
+            }
 
-        if util.is_set(obj):
-            if not self.unpicklable:
-                return list_recurse
-            return lambda obj: {tags.SET: [self._flatten(v) for v in obj]}
-
-        if util.is_dictionary(obj):
-            return self._flatten_dict_obj
-
-        if util.is_type(obj):
-            return _mktyperef
-
-        if util.is_object(obj):
+        elif util.is_object(obj):
             return self._ref_obj_instance
 
-        if util.is_module_function(obj):
+        elif util.is_type(obj):
+            return _mktyperef
+
+        elif util.is_module_function(obj):
             return self._flatten_function
 
         # instance methods, lambdas, old style classes...
@@ -378,20 +432,13 @@ class Pickler(object):
             self._mkref(obj)
             return self._flatten_obj_instance(obj)
 
-    def _flatten_file(self, obj):
-        """
-        Special case file objects
-        """
-        assert not PY3 and isinstance(obj, types.FileType)
-        return None
-
-    def _flatten_bytestring(self, obj):
-        if PY2:
-            try:
-                return obj.decode('utf-8')
-            except UnicodeDecodeError:
-                pass
-        return {self._bytes_tag: self._bytes_encoder(obj)}
+    def _getstate(self, obj, data):
+        state = self._flatten(obj)
+        if self.unpicklable:
+            data[tags.STATE] = state
+        else:
+            data = state
+        return data
 
     def _flatten_obj_instance(self, obj):
         """Recursively flatten an instance and return a json-friendly dict"""
@@ -558,12 +605,47 @@ class Pickler(object):
         self._pickle_warning(obj)
         return None
 
-    def _flatten_function(self, obj):
-        if self.unpicklable:
-            data = {tags.FUNCTION: util.importable_name(obj)}
-        else:
-            data = None
+    def _escape_key(self, k):
+        return tags.JSON_KEY + encode(
+            k,
+            reset=False,
+            keys=True,
+            context=self,
+            backend=self.backend,
+            make_refs=self.make_refs,
+        )
 
+    def _flatten_non_string_key_value_pair(self, k, v, data):
+        """Flatten only non-string key/value pairs"""
+        if not util.is_picklable(k, v):
+            return data
+        if self.keys and not isinstance(k, string_types):
+            k = self._escape_key(k)
+            data[k] = self._flatten(v)
+        return data
+
+    def _flatten_string_key_value_pair(self, k, v, data):
+        """Flatten string key/value pairs only."""
+        if not util.is_picklable(k, v):
+            return data
+        if self.keys:
+            if not isinstance(k, string_types):
+                return data
+            elif k.startswith(tags.JSON_KEY):
+                k = self._escape_key(k)
+        else:
+            if k is None:
+                k = 'null'  # for compatibility with common json encoders
+
+            if self.numeric_keys and isinstance(k, numeric_types):
+                pass
+            elif not isinstance(k, string_types):
+                try:
+                    k = repr(k)
+                except Exception:
+                    k = compat.ustr(k)
+
+        data[k] = self._flatten(v)
         return data
 
     def _flatten_dict_obj(self, obj, data=None):
@@ -601,7 +683,7 @@ class Pickler(object):
                 if self._mkref(factory):
                     # We've never seen this object before so pickle it in-place.
                     # Create an instance from the factory and assume that the
-                    # resulting instance is a suitable examplar.
+                    # resulting instance is a suitable exemplar.
                     value = self._flatten_obj_instance(handlers.CloneFactory(factory()))
                 else:
                     # We've seen this object before.
@@ -622,7 +704,11 @@ class Pickler(object):
         ok = False
         for k in attrs:
             try:
-                value = getattr(obj, k)
+                if not k.startswith('__'):
+                    value = getattr(obj, k)
+                else:
+                    # we can use f-strings once we drop < 3.7 in jsonpickle 3.0
+                    value = getattr(obj, "_" + obj.__class__.__name__ + k)
                 flatten(k, value, data)
             except AttributeError:
                 # The attribute may have been deleted
@@ -664,39 +750,6 @@ class Pickler(object):
         data[k] = self._flatten(v)
         return data
 
-    def _flatten_non_string_key_value_pair(self, k, v, data):
-        """Flatten only non-string key/value pairs"""
-        if not util.is_picklable(k, v):
-            return data
-        if self.keys and not isinstance(k, string_types):
-            k = self._escape_key(k)
-            data[k] = self._flatten(v)
-        return data
-
-    def _flatten_string_key_value_pair(self, k, v, data):
-        """Flatten string key/value pairs only."""
-        if not util.is_picklable(k, v):
-            return data
-        if self.keys:
-            if not isinstance(k, string_types):
-                return data
-            elif k.startswith(tags.JSON_KEY):
-                k = self._escape_key(k)
-        else:
-            if k is None:
-                k = 'null'  # for compatibility with common json encoders
-
-            if self.numeric_keys and isinstance(k, numeric_types):
-                pass
-            elif not isinstance(k, string_types):
-                try:
-                    k = repr(k)
-                except Exception:
-                    k = compat.ustr(k)
-
-        data[k] = self._flatten(v)
-        return data
-
     def _flatten_sequence_obj(self, obj, data):
         """Return a json-friendly dict for a sequence subclass."""
         if hasattr(obj, '__dict__'):
@@ -707,52 +760,3 @@ class Pickler(object):
         else:
             return value
         return data
-
-    def _escape_key(self, k):
-        return tags.JSON_KEY + encode(
-            k,
-            reset=False,
-            keys=True,
-            context=self,
-            backend=self.backend,
-            make_refs=self.make_refs,
-        )
-
-    def _getstate(self, obj, data):
-        state = self._flatten(obj)
-        if self.unpicklable:
-            data[tags.STATE] = state
-        else:
-            data = state
-        return data
-
-    def _pickle_warning(self, obj):
-        if self.warn:
-            msg = 'jsonpickle cannot pickle %r: replaced with None' % obj
-            warnings.warn(msg)
-
-
-def _in_cycle(obj, objs, max_reached, make_refs):
-    """Detect cyclic structures that would lead to infinite recursion"""
-    return (
-        (max_reached or (not make_refs and id(obj) in objs))
-        and not util.is_primitive(obj)
-        and not util.is_enum(obj)
-    )
-
-
-def _mktyperef(obj):
-    """Return a typeref dictionary
-
-    >>> _mktyperef(AssertionError) == {'py/type': 'builtins.AssertionError'}
-    True
-
-    """
-    return {tags.TYPE: util.importable_name(obj)}
-
-
-def _wrap_string_slot(string):
-    """Converts __slots__ = 'a' into __slots__ = ('a',)"""
-    if isinstance(string, string_types):
-        return (string,)
-    return string
